c  This is a simple 3-Body code using the Log-Hamiltonian method.
c  The triangle sides are integrated (like in Heggie's method).
c
c  Output: standard output: error info; 
c          file fort.66 time, coords;
c          file fort.67 time, cm-coords

c An (example) data fil

c                     100.   .01    1.e-13   /tmx, tincr, eps
c                     1  0 0 0   0.1 -0.01 0.02       /m x  y z vx vy vz
c                     1.1 1 0 0   0.03 -.0010 0.01      /    -"-
c                     1.2 0 1 0   1.e-1 1.e-2 1.e-4      /    -"-


         implicit real*8 (a-h,m,o-z)
         real*8 xr(9),vr(9),xw(9)
         real*8 g0(3),gt(3),cmx(3),cmv(3)
         common/diagno/neval
         common/masses/mass,m(3),mu(3),mm(3)
         common/kinpot/TKIN,UPOT,Energy
         open(66,file='xyz.dat')
         open(67,file='xyz.cm')
         neval=0
         read(5,*)tmx,tincr,eps ! read max time, output interval and tolerance
         do i=1,3
         k1=i*3-2
         k2=k1+2
         read(5,*)m(i),(xr(k),k=k1,k2),(vr(k),k=k1,k2) ! read mass & initial coord/vels
         end do
         nb=3
         newcase=1 ! *newer*  set this=1 in the middle of an integaration 
      !                          (only initially, when starting a new case)
         time=0
         call CMsys(m,xr,cmx) ! obtain cm-coords  
         call CMsys(m,vr,cmv) ! obtain cm-vels
         call Constants Of Motion(xr,vr,m,nb,ene0,g0,alag) ! initial const. 
c       
1        continue         
         call ThreeB(m,time,xr,vr,tincr,newcase,eps) !integrate over time-interval = tincr 
         call Constants Of Motion(xr,vr,m,nb,enet,gt,alag) ! constants for checking
         vrh=abs(enet-ene0)/alag
         glag=mass**2.5/abs(ene0)*sqrt(alag)
         dgg=(gt(1)-g0(1))**2+(gt(2)-g0(2))**2+(gt(3)-g0(3))**2
         dgg=sqrt(dgg)/glag ! a measure for angular mom.error
         write(6,106)time,vrh,dgg,neval! write errors & # function calls
106      format(1x,f12.4,1p,2g11.2,i7,' t, dE/L, dam, Neval')
         do i=1,3
         i0=3*(i-1)
         do k=1,3
         xw(i0+k)=xr(i0+k)+cmx(k)+cmv(k)*time ! add cm-motion
         end do
         end do
         write(66,123)time,xw! print coordinates
         write(67,123)time,xr! print cm-coords (vels are in vr(.))
123      format(1x,f12.4,1p,9g12.4)
         call flush(6)
         call flush(66)
         if(time.lt.tmx)goto 1
         end

         subroutine ThreeB(mr,time,xr,vr,tincr,newcase,eps)
         implicit real*8 (a-h,m,o-z)
         real*8 xr(9),vr(9),x(0:9),v(0:9),y(20),SY(20),mr(3)
         common/masses/mass,m(3),mu(3),mm(3)
         common/kinpot/TKIN,UPOT,Energy
         equivalence(y(1),x(0)),(y(11),v(0))
         save
c        initializations         
         if(newcase.eq.1)then
         newcase=0
         m(1)=mr(1)
         m(2)=mr(2)
         m(3)=mr(3)
         mass=m(1)+m(2)+m(3)
         mu(1)=m(1)/mass
         mu(2)=m(2)/mass
         mu(3)=m(3)/mass
         mm(1)=m(2)*m(3)
         mm(2)=m(1)*m(3)
         mm(3)=m(1)*m(2)
         call Compute Triangle Differences(xr(1),x(1)) !coords
         call Compute Triangle Differences(vr(1),v(1)) !vels
         call Evaluate Energy(v,x,E)
         h=.01/(TKIN+UPOT) 
         Energy=E ! to common
         v(0)=-E ! binding energy
         x(0)=0  ! time
         end if ! newcase
c         
         told=x(0)
         nsteps=0
         NVX=20

1        continue
         h=min(h,tincr*Upot) ! comment out if you want it to run as fast as it can
         call  DIFSYAB(NVX,EPS,SY,h,dummy,Y) ! essentially =DIFSY2
         call Triangle Correction(v,x)

                 if(h.eq.0.0)then
                 write(6,*)' STEP=0!',char(7),' MUST STOP!'
                 STOP
                 end if

         nsteps=nsteps+1 !         to avoid infinite loop
         test=x(0)+.5*min(h,tincr*Upot)/Upot
         if(test.lt.told+tincr.and.nsteps.lt.10000000)goto 1
         time=x(0)


         call Transform to CM system(X(1),XR(1)) ! cm coords
         call Transform to CM system(V(1),VR(1)) ! cm vels

         return
         end
         subroutine Triangle Correction(v,x)
         implicit real*8 (a-h,o-z)
         real*8 v(0:9),x(0:9)
c        adjust such that vector sums are==0
c        find longest distance
         save
         rrmx=0
         do i=1,3
         i1=3*i-2
         rr=sqa(x(i1))
         if(rrmx.lt.rr)then
         rrmx=rr
         imx=i
         end if
         end do ! k

         ix0=3*imx-3
c        take =sum of two shortest for the longest
         do k=1,3
         x(ix0+k)=0
         v(ix0+k)=0
         end do

         do i=1,3
         i0=3*i-3
         if(i.ne.imx)then
         do k=1,3
         x(ix0+k)=x(ix0+k)-x(i0+k) ! re-evaluate the longest vector
         v(ix0+k)=v(ix0+k)-v(i0+k) ! and the velocity in the same way
         end do
         end if
         end do
         return
         end 
         
         subroutine SubSteps(Y,H,NL) ! NL substeps of size H
         implicit real*8 (a-h,o-z)
         real*8 y(20)
         call LeapFrog(y(1),y(11),H,NL)
         return
         end
         subroutine LeapFrog(x,v,h,Leaps)
         implicit real*8 (a-h,m,o-z)
         real*8 x(0:9),v(0:9)
         save
         call Move Coordinates(v,x,h/2)
         call Move Velocities(v,x,h)
         do L=1,Leaps-1
         call Move Coordinates(v,x,h)
         call Move Velocities(v,x,h)
         end do
         call Move Coordinates(v,x,h/2)
         return
         end
         subroutine Evaluate Energy(v,x,E)
         implicit real*8 (a-h,m,o-z)
         real*8 x(0:9),v(0:9)
         common/masses/mass,m(3),mu(3),mm(3)
         common/kinpot/T,U,Energy
         T=0
         U=0
         do i=1,3
         i1=3*i-2
         vv=sqa(v(i1))
         rr=sqa(x(i1))
         R=sqrt(rr)
         T=T+mm(i)*vv/(2*mass)
         U=U+mm(i)/R
         end do
         E=T-U
         return
         end
         subroutine Transform to CM system(X,Y)
         implicit real*8 (a-h,m,o-z)
         real*8 Y(9),X(9)
         parameter(I1=0,I2=3,I3=6)
         common/masses/mass,m(3),mu(3),mm(3)
         do k=1,3
         Y(I1+k)=X(I2+k)*mu(3)-X(I3+k)*mu(2)
         Y(I2+k)=X(I3+k)*mu(1)-X(I1+k)*mu(3)
         Y(I3+k)=X(I1+k)*mu(2)-X(I2+k)*mu(1)
         end do 
         return
         end
         subroutine Compute Triangle Differences(X,Y)
         real*8 Y(9),x(9)
         parameter(I1=0,I2=3,I3=6)
         do k=1,3
         Y(I1+k)=X(I3+k)-X(I2+k)
         Y(I2+k)=X(I1+k)-X(I3+k)
         Y(I3+k)=X(I2+k)-X(I1+k)
         end do
         return
         end 
         subroutine Move Coordinates(v,x,s)
         implicit real*8 (a-h,m,o-z)
         real*8 x(0:9),v(0:9)
         common/masses/mass,m(3),mu(3),mm(3)
         Te=v(0)
         do k=1,3
         Te=Te+mm(k)*sqa(v(3*k-2))/(2*mass) ! kinetic energy
         end do
         dt=s/Te                 ! time increment
         x(0)=x(0)+dt            ! add it 
         do k=1,9
         x(k)=x(k)+dt*v(k)       ! coords move linearly
         end do
         return
         end
         function sqa(a)
         real*8 a(3),sqa
         sqa=a(1)**2+a(2)**2+a(3)**2
         return
         end
         subroutine Move Velocities(v,x,s)
         implicit real*8 (a-h,m,o-z)
         parameter(I1=0,I2=3,I3=6)
         real*8 x(0:9),v(0:9),A(9),SumA(3)
         common/masses/mass,m(3),mu(3),mm(3)
         common/diagno/neval
         common/kinpot/Tkin,U,Energy
         save
                               tiny=1.d-30
         neval=neval+1
         R1=sqrt(sqa(x(1)))! +tiny ! distances
         R2=sqrt(sqa(x(4)))! +tiny
         R3=sqrt(sqa(x(7)))! +tiny  
         U=mm(1)/R1+mm(2)/R2+mm(3)/R3 !potential
         do k=1,3
         A(I1+k)=x(I1+k)/R1**3 ! acc-terms
         A(I2+k)=x(I2+k)/R2**3
         A(I3+k)=x(I3+k)/R3**3
         SumA(k)=A(I1+k)+A(I2+k)+A(I3+k)
         end do
         
         dt=s/U
         do k=1,3
         v(I1+k)=v(I1+k)+dt*(-mass*A(I1+k)+m(1)*SumA(k) ) !velocity jumps
         v(I2+k)=v(I2+k)+dt*(-mass*A(I2+k)+m(2)*SumA(k) )  
         v(I3+k)=v(I3+k)+dt*(-mass*A(I3+k)+m(3)*SumA(k) )  
         end do
         return
         end

        SUBROUTINE DIFSYAB(N,EPS,S,h,t,Y)
        implicit real*8 (a-h,o-z)
        parameter (NMX=20,NMX2=2*NMX,nmx27=nmx2*7) ! NMX=MAX(N),N=3*NB
        REAL*8 Y(N),YR(NMX2),YS(NMX2)
     +  ,DT(NMX2,7),D(7),S(N),EP(4)
        LOGICAL KONV,BO,KL,GR
        DATA EP/.4D-1,.16D-2,.64D-4,.256D-5/
        data dt/nmx27*0.0d0/
        save
c------------------------------
        old_time=y(1)  ! improve checking
                 y(1)=0 ! of the time
c------------------------------

        IF(EPS.LT.1.D-15)EPS=1.D-15
        IF(N.gt.NMX)write(6,*) ' too many variables!',char(7),'MAX=',NMX
        JTI=0
        FY=1
        redu=0.7d0
10      tN=t+H
        BO=.FALSE.
C
        M=1
        JR=2
        JS=3

        DO 260 J=1,5!10

        do i=1,N
        ys(i)=y(i)
c        s(i)=max(abs(ys(i)),s(i)) 
        end do
C

        IF(BO)then
        D(2)=1.777777777777778D0
        D(4)=7.111111111111111D0
        D(6)=28.44444444444444D0
        else
        D(2)=2.25D0
        D(4)=9.D0
        D(6)=3.6D1
        end if

        IF(J.gt.7)then
        L=7
        D(7)=6.4D1
        else
        L=J
        D(L)=M*M
        end if

        KONV=L.GT.3
c----------------------------------------------------------------
           subH=H/M
           call SubSteps(YS,subH,M) ! M substeps of size H/M.
c         mean derivatives for { dLambda/dQ* (delta Q)<eps } -test
          Nh=N/2
          do i=1,NH
          Si=abs(YS(I+Nh)-Y(I+Nh))/H
          if(si.ne.0.0)S(I)=max(S(I),1/SI)
          SI=abs(YS(I)-Y(I))/H
          if(si.ne.0.0)S(I+Nh)=max(S(I+Nh),1/SI)
          end do
c----------------------------------------------------------------
        KL=L.LT.2
        GR=L.GT.5
        FS=0.

        DO 233 I=1,N 
        V=DT(I,1)
        C=YS(I)
        DT(I,1)=C
        TA=C
        IF(KL)GO TO 233
        DO  K=2,L
        B1=D(K)*V
        B=B1-C
        W=C-V
        U=V
        if(B.ne.0.0)then
        B=W/B
        U=C*B
        C=B1*B
        end if
        V=DT(I,K)
        DT(I,K)=U
        TA=U+TA
        END DO ! K=2,L
        SI=max(S(I),abs(TA))
         S(I)=SI ! not needed
        IF(DABS(YR(I)-TA).GT.SI*EPS) KONV=.FALSE.
        IF(GR.OR.SI.EQ.0.D0)GO TO 233
        FV=DABS(W)/SI
        IF(FS.LT.FV)FS=FV
233     YR(I)=TA
c       end of I-loop
        IF(FS.EQ.0.D0)GO TO 250
        FA=FY
        K=L-1
        FY=(EP(K)/FS)**(1.d0/FLOAT(L+K))
        fy=min(fy,1.4d0)
        IF(L.EQ.2)GO TO 240
        IF(FY.LT.0.7*FA)GO TO 250
240     IF(FY.GT.0.7)GO TO 250
        H=H*FY
        JTI=JTI+1
        IF(JTI.GT.50)GO TO 30
        GO TO 10
250     IF(KONV)GO TO 20
        D(3)=4.D0
        D(5)=1.6D1
        BO=.NOT.BO
        M=JR
        JR=JS
260     JS=M+M
        redu=redu*redu+.001d0
        H=H*redu !  .5D0
        GO TO 10
20      t=tN
        H=H*FY
        DO  I=1,N
        Y(I)=YR(I)
        END DO
        Y(1)=Y(1)+old_time
        RETURN
30      H=0.
        Y(1)=Y(1)+old_time
        RETURN
        END

        subroutine CMsys(m,x,cx)
        implicit real*8 (a-h,m,o-z)
        real*8 x(9),cx(3),m(3)
        mass=m(1)+m(2)+m(3)
        do k=1,3
        cx(k)=0
        do i=1,3
        cx(k)=cx(k)+m(i)*x(k+3*i-3)
        end do
        cx(k)=cx(k)/mass
        end do 
        do i=1,3
        i0=3*i-3
        do k=1,3
        x(i0+k)=x(i0+k)-cx(k)
        end do
        end do
        return
        end
        subroutine constants of motion(x,xd,m,nb,energy,g,alag)
        implicit real*8 (a-h,m,o-z)
        dimension x(*),xd(*),m(*),g(3)
        save
        t=0.0
        ug=0.0
        u=ug
        g(1)=0.
        g(2)=0.
        g(3)=0.
        rmin=1.d30
        do  i=1,nb
        k1=(i-1)*3+1
        k2=k1+1
        k3=k2+1
        t=t+.5d0*m(i)*(xd(k1)**2+xd(k2)**2+xd(k3)**2)
        g(1)=g(1)+m(i)*(x(k2)*xd(k3)-x(k3)*xd(k2))
        g(2)=g(2)-m(i)*(x(k1)*xd(k3)-x(k3)*xd(k1))
        g(3)=g(3)+m(i)*(x(k1)*xd(k2)-x(k2)*xd(k1))
        if(i.lt.nb)then
        j1=i+1
        do  j=j1,nb
        ki=(i-1)*3
        kj=(j-1)*3
        r2=0.
        do  k=1,3
        ki=ki+1
        kj=kj+1
        r2=r2+(x(ki)-x(kj))**2
        end do
        u=u+m(i)*m(j)/sqrt(r2)
        end do
        end if
        end do
        energy=t-u
        alag=t+u
        return
        end
